# Copyright (c) 2024-2025, Kedar Hippalgaonkar's Materials by Design Lab
# Distributed under the terms of the MIT License.

"""
Virp (VIRtual cell generation by Permutation) is a code for generating a virtual cell from a site-disordered .cif crystal file
"""

from .main import *
from .enumerate import *
from .relax import *

__copyright__  = "Kedar Hippalgaonkar's Materials by Design Lab"
__version__    =  "1.3.1"
__maintainer__ = "Andy Paul Chen"
__email__      = "la.vache.qui.vit@gmail.com"
__status__     = "Development"
__date__       = "26 May 2025"